from scipy.ndimage.filters import convolve
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
from itertools import groupby
from numpy import linalg as LA

def Hessian2D(I, Sigma=1):
    X, Y = np.meshgrid(np.arange(-np.round(3*Sigma), np.round(3*Sigma) +1),
                       np.arange(-np.round(3*Sigma), np.round(3*Sigma) +1), indexing='ij')
    DGaussxx = 1/(2*np.pi*Sigma**4)*(X**2/Sigma**2 - 1)*np.exp(-(X**2 + Y**2)/(2*Sigma**2))
    DGaussxy = (1/(2*np.pi*Sigma**6))*(X*Y)*np.exp(-(X**2 + Y**2)/(2*Sigma**2))
    DGaussyy = DGaussxx.conj().T
    Dxx = convolve(I, DGaussxx, mode='constant', cval=0.0)
    Dxy = convolve(I, DGaussxy, mode='constant', cval=0.0)
    Dyy = convolve(I, DGaussyy, mode='constant', cval=0.0)
    
    return Dxx, Dxy, Dyy

def eig2image(Dxx, Dxy, Dyy):
    
    tmp = np.sqrt((Dxx - Dyy)**2 + 4*Dxy**2)
    v2x = 2*Dxy
    v2y = Dyy - Dxx + tmp
    
    # Normalize
    mag = np.sqrt(v2x**2 + v2y**2)
    i = np.invert(np.isclose(mag, np.zeros(mag.shape)))
    v2x[i] = v2x[i]/mag[i]
    v2y[i] = v2y[i]/mag[i]
    
    # The eigenvectors are orthogonal
    v1x = -v2y.copy()
    v1y = v2x.copy()
    
    # Compute the eigenvalues
    mu1 = 0.5*(Dxx + Dyy + tmp)
    mu2 = 0.5*(Dxx + Dyy - tmp)
    
    
    # sorted ascending
    return mu2, mu1, v1x, v1y


def VesselFilter2D(I, FrangiScaleRange=np.array([1, 5]), FrangiScaleRatio=0.5,BlackWhite=True,verbose=False):
    if len(FrangiScaleRange) > 1:
        sigmas = np.arange(FrangiScaleRange[0], FrangiScaleRange[1]+1, FrangiScaleRatio)
        sigmas = sorted(sigmas)
    else:
        sigmas = [FrangiScaleRange[0]]
    
    # Make matrices to store all filterd images
    Lambda1 = np.zeros([I.shape[0], I.shape[1], len(sigmas)])
    Lambda2 = np.zeros([I.shape[0], I.shape[1], len(sigmas)])
    ALLangles = np.zeros([I.shape[0], I.shape[1], len(sigmas)])
    
    for i in range(len(sigmas)):
        if verbose: 
            print('Current Frangi Filter Sigma: ', str(sigmas[i]))
            
            
        Dxx, Dxy, Dyy = Hessian2D(I, sigmas[i])
        
        # Correct for scale
        Dxx *= (sigmas[i]**2)
        Dxy *= (sigmas[i]**2)
        Dyy *= (sigmas[i]**2)
        
        if BlackWhite:
            Lambda1[:,:,i], Lambda2[:,:,i], Ix, Iy = eig2image(Dxx, Dxy, Dyy)
            ALLangles[:,:,i] = np.arctan2(Ix, Iy)
        else:
            Lambda2[:,:,i], Lambda1[:,:,i], Ix, Iy = eig2image(Dxx, Dxy, Dyy)
            Lambda1[:,:,i]/=-1.
            Lambda2[:,:,i]/=-1.
            ALLangles[:,:,i] = np.arctan2(-Iy, Ix)
    
        # compute the best response and the corresponding scale for each point
    if len(sigmas) > 1:
        outIm = np.amax(Lambda2, axis=2)
        outIm = outIm.reshape(I.shape[0], I.shape[1], order='F')
        whatScale = np.argmax(Lambda2, axis=2)
        whatScale = np.reshape(whatScale, I.shape, order='F')
    else:
        outIm = ALLfiltered.reshape(I.shape[0], I.shape[1], order='F')
        whatScale = np.ones(I.shape)
                
    return Lambda1, Lambda2, ALLangles, outIm, whatScale

def AnisotropicLiftedRiemannianMetric(alpha_,beta_,lambda1_,lambda2_,oriens_):
    v1=np.cos(oriens_)
    v2=np.sin(oriens_)
    p1=np.exp(alpha_*lambda2_)
    p2=np.exp(alpha_*lambda1_)
    metric11=p1*v1**2+p2*v2**2
    metric12=(p1-p2)*v1*v2
    metric22=p1*v2**2+p2*v1**2
    metric31=np.zeros(v1.shape);
    metric32=np.zeros(v1.shape);
    metric33=beta_*p1
    metric=np.stack((metric11,metric12,metric22,metric31,metric32,metric33),3);
    return metric
